@@ -37,6 +37,7 @@
 #include <getopt.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
+#include <assert.h>
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/hci.h>
@@ -57,14 +58,14 @@
 #include "telephony.h"
 #include "headset.h"
 #include "glib-helper.h"
+#include "dbus-common.h"
+#include "../src/adapter.h"
+#include "../src/device.h"
 
 #define DC_TIMEOUT 3000
 
 #define RING_INTERVAL 3000
 
-/* Number of indicator events that can be queued */
-#define EV_BUF_SIZE 4
-
 #define BUF_SIZE 1024
 
 #define HEADSET_GAIN_SPEAKER 'S'
@@ -77,24 +78,24 @@ static struct {
 	int er_mode;			/* Event reporting mode */
 	int er_ind;			/* Event reporting for indicators */
 	int rh;				/* Response and Hold state */
-	gboolean ev_buf_active;		/* Buffer indicator events */
-	struct {
-		int index;		/* HFP indicator index */
-		int val;		/* new indicator value */
-	} ev_buf[EV_BUF_SIZE];		/* Indicator event buffer */
+	char *number;			/* Incoming phone number */
+	int number_type;		/* Incoming number type */
+	guint ring_timer;		/* For incoming call indication */
+	const char *chld;		/* Response to AT+CHLD=? */
 } ag = {
 	.telephony_ready = FALSE,
 	.features = 0,
 	.er_mode = 3,
 	.er_ind = 0,
 	.rh = -1,
-	.ev_buf_active = FALSE,
-	.ev_buf = { { 0, 0 } },
+	.number = NULL,
+	.number_type = 0,
+	.ring_timer = 0,
 };
 
 static gboolean sco_hci = TRUE;
 
-static struct audio_device *active_telephony_device = NULL;
+static GSList *active_devices = NULL;
 
 static char *str_state[] = {
 	"HEADSET_STATE_DISCONNECTED",
@@ -132,8 +133,6 @@ struct headset {
 
 	gboolean auto_dc;
 
-	guint ring_timer;
-
 	guint dc_timer;
 
 	char buf[BUF_SIZE];
@@ -143,8 +142,8 @@ struct headset {
 	gboolean hfp_active;
 	gboolean search_hfp;
 	gboolean cli_active;
-	char *ph_number;
-	int type;
+	gboolean cme_enabled;
+	gboolean cwa_enabled;
 
 	headset_state_t state;
 	struct pending_connect *pending;
@@ -161,6 +160,12 @@ struct event {
 	int (*callback) (struct audio_device *device, const char *buf);
 };
 
+static inline DBusMessage *invalid_args(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
+			"Invalid arguments in method call");
+}
+
 static DBusHandlerResult error_not_supported(DBusConnection *conn,
 							DBusMessage *msg)
 {
@@ -253,16 +258,13 @@ static void print_hf_features(uint32_t features)
 	g_free(str);
 }
 
-static int headset_send(struct headset *hs, char *format, ...)
+static int headset_send_valist(struct headset *hs, char *format, va_list ap)
 {
 	char rsp[BUF_SIZE];
-	va_list ap;
 	ssize_t total_written, written, count;
 	int fd;
 
-	va_start(ap, format);
 	count = vsnprintf(rsp, sizeof(rsp), format, ap);
-	va_end(ap);
 
 	if (count < 0)
 		return -EINVAL;
@@ -286,47 +288,16 @@ static int headset_send(struct headset *hs, char *format, ...)
 	return 0;
 }
 
-static int buffer_event(int index)
+static int headset_send(struct headset *hs, char *format, ...)
 {
-	int i;
+	va_list ap;
+	int ret;
 
-	for (i = 0; i < EV_BUF_SIZE; i++) {
-		if (ag.ev_buf[i].index == 0) {
-			ag.ev_buf[i].index = index + 1;
-			ag.ev_buf[i].val = ag.indicators[index].val;
-			return 0;
-		}
-	}
+	va_start(ap, format);
+	ret = headset_send_valist(hs, format, ap);
+	va_end(ap);
 
-	error("No space in event buffer");
-	return -ENOSPC;
-}
-
-static int flush_events(void)
-{
-	int i;
-	struct headset *hs;
-
-	if (!active_telephony_device)
-		return -ENODEV;
-
-	hs = active_telephony_device->headset;
-
-	for (i = 0; i < EV_BUF_SIZE; i++) {
-		int ret;
-
-		if (ag.ev_buf[i].index == 0)
-			break;
-
-		ret = headset_send(hs, "\r\n+CIEV:%d,%d\r\n",
-					ag.ev_buf[i].index, ag.ev_buf[i].val);
-		if (ret < 0)
-			return ret;
-
-		ag.ev_buf[i].index = 0;
-	}
-
-	return 0;
+	return ret;
 }
 
 static int supported_features(struct audio_device *device, const char *buf)
@@ -559,6 +530,41 @@ static int sco_connect(struct audio_device *dev, headset_stream_cb_t cb,
 	return 0;
 }
 
+static int hfp_cmp(struct headset *hs)
+{
+	if (hs->hfp_active)
+		return 0;
+	else
+		return -1;
+}
+
+static void send_foreach_headset(GSList *devices,
+					int (*cmp)(struct headset *hs),
+					char *format, ...)
+{
+	GSList *l;
+	va_list ap;
+
+	for (l = devices; l != NULL; l = l->next) {
+		struct audio_device *device = l->data;
+		struct headset *hs = device->headset;
+		int ret;
+
+		assert(hs != NULL);
+
+		if (cmp && cmp(hs) != 0)
+			continue;
+
+		va_start(ap, format);
+		ret = headset_send_valist(hs, format, ap);
+		if (ret < 0)
+			error("Failed to send to headset: %s (%d)",
+					strerror(-ret), -ret);
+		va_end(ap);
+	}
+}
+
+
 static void hfp_slc_complete(struct audio_device *dev)
 {
 	struct headset *hs = dev->headset;
@@ -571,26 +577,64 @@ static void hfp_slc_complete(struct audio_device *dev)
 	if (p == NULL)
 		return;
 
-	if (p->msg) {
-		DBusMessage *reply = dbus_message_new_method_return(p->msg);
-		dbus_connection_send(dev->conn, reply, NULL);
-		dbus_message_unref(reply);
-	}
-
 	if (p->target_state == HEADSET_STATE_CONNECTED) {
+		if (p->msg) {
+			DBusMessage *reply = dbus_message_new_method_return(p->msg);
+			dbus_connection_send(dev->conn, reply, NULL);
+			dbus_message_unref(reply);
+		}
 		pending_connect_finalize(dev);
 		return;
 	}
 
 	p->err = sco_connect(dev, NULL, NULL, NULL);
-	if (p->err < 0)
+	if (p->err < 0) {
+		if (p->msg)
+			error_connection_attempt_failed(dev->conn, p->msg, p->err);
 		pending_connect_finalize(dev);
+	}
+}
+
+static int telephony_generic_rsp(struct audio_device *device, cme_error_t err)
+{
+	struct headset *hs = device->headset;
+
+	if (err != CME_ERROR_NONE) {
+		if (hs->cme_enabled)
+			return headset_send(hs, "\r\n+CME ERROR:%d\r\n", err);
+		else
+			return headset_send(hs, "\r\nERROR\r\n");
+	}
+
+	return headset_send(hs, "\r\nOK\r\n");
+}
+
+int telephony_event_reporting_rsp(void *telephony_device, cme_error_t err)
+{
+	struct audio_device *device = telephony_device;
+	struct headset *hs = device->headset;
+	int ret;
+
+	if (err != CME_ERROR_NONE)
+		return telephony_generic_rsp(telephony_device, err);
+
+	ret = headset_send(hs, "\r\nOK\r\n");
+	if (ret < 0)
+		return ret;
+
+	if (hs->state != HEADSET_STATE_CONNECT_IN_PROGRESS)
+		return 0;
+
+	if (ag.features & AG_FEATURE_THREE_WAY_CALLING)
+		return 0;
+
+	hfp_slc_complete(device);
+
+	return 0;
 }
 
 static int event_reporting(struct audio_device *dev, const char *buf)
 {
-	struct headset *hs = dev->headset;
-	int ret;
 	char **tokens; /* <mode>, <keyp>, <disp>, <ind>, <bfr> */
 
 	if (strlen(buf) < 13)
@@ -614,24 +658,12 @@ static int event_reporting(struct audio_device *dev, const char *buf)
 	switch (ag.er_ind) {
 	case 0:
 	case 1:
-		telephony_event_reporting_req(ag.er_ind);
+		telephony_event_reporting_req(dev, ag.er_ind);
 		break;
 	default:
 		return -EINVAL;
 	}
 
-	ret = headset_send(hs, "\r\nOK\r\n");
-	if (ret < 0)
-		return ret;
-
-	if (hs->state != HEADSET_STATE_CONNECT_IN_PROGRESS)
-		return 0;
-
-	if (ag.features & AG_FEATURE_THREE_WAY_CALLING)
-		return 0;
-
-	hfp_slc_complete(dev);
-
 	return 0;
 }
 
@@ -640,7 +672,15 @@ static int call_hold(struct audio_device *dev, const char *buf)
 	struct headset *hs = dev->headset;
 	int err;
 
-	err = headset_send(hs, "\r\n+CHLD:(0,1,1x,2,2x,3,4)\r\n");
+	if (strlen(buf) < 9)
+		return -EINVAL;
+
+	if (buf[8] != '?') {
+		telephony_call_hold_req(dev, &buf[8]);
+		return 0;
+	}
+
+	err = headset_send(hs, "\r\n+CHLD:(%s)\r\n", ag.chld);
 	if (err < 0)
 		return err;
 
@@ -664,36 +704,68 @@ static int button_press(struct audio_device *device, const char *buf)
 			AUDIO_HEADSET_INTERFACE, "AnswerRequested",
 			DBUS_TYPE_INVALID);
 
-	if (hs->ring_timer) {
-		g_source_remove(hs->ring_timer);
-		hs->ring_timer = 0;
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
 	}
 
 	return headset_send(hs, "\r\nOK\r\n");
 }
 
-static int answer_call(struct audio_device *device, const char *buf)
+int telephony_answer_call_rsp(void *telephony_device, cme_error_t err)
 {
+	struct audio_device *device = telephony_device;
 	struct headset *hs = device->headset;
 
-	ag.ev_buf_active = TRUE;
+	if (err != CME_ERROR_NONE)
+		return telephony_generic_rsp(telephony_device, err);
 
-	if (telephony_answer_call_req() < 0) {
-		headset_send(hs, "\r\nERROR\r\n");
-		return 0;
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
 	}
 
-	flush_events();
-	ag.ev_buf_active = FALSE;
-
-	if (hs->ring_timer) {
-		g_source_remove(hs->ring_timer);
-		hs->ring_timer = 0;
+	if (ag.number) {
+		g_free(ag.number);
+		ag.number = NULL;
 	}
 
-	if (hs->ph_number) {
-		g_free(hs->ph_number);
-		hs->ph_number = NULL;
+	return headset_send(hs, "\r\nOK\r\n");
+}
+
+int telephony_anser_call_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int answer_call(struct audio_device *device, const char *buf)
+{
+	telephony_answer_call_req(device);
+
+	return 0;
+}
+
+int telephony_terminate_call_rsp(void *telephony_device,
+					cme_error_t err)
+{
+	struct audio_device *device = telephony_device;
+	struct headset *hs = device->headset;
+
+	if (err != CME_ERROR_NONE)
+		return telephony_generic_rsp(telephony_device, err);
+
+	g_dbus_emit_signal(device->conn, device->path,
+			AUDIO_HEADSET_INTERFACE, "CallTerminated",
+			DBUS_TYPE_INVALID);
+
+	if (ag.number) {
+		g_free(ag.number);
+		ag.number = NULL;
+	}
+
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
 	}
 
 	return headset_send(hs, "\r\nOK\r\n");
@@ -701,33 +773,9 @@ static int answer_call(struct audio_device *device, const char *buf)
 
 static int terminate_call(struct audio_device *device, const char *buf)
 {
-	struct headset *hs = device->headset;
+	telephony_terminate_call_req(device);
 
-	ag.ev_buf_active = TRUE;
-
-	if (telephony_terminate_call_req() < 0) {
-		headset_send(hs, "\r\nERROR\r\n");
-		return 0;
-	}
-
-	flush_events();
-	ag.ev_buf_active = FALSE;
-
-	g_dbus_emit_signal(device->conn, device->path,
-			AUDIO_HEADSET_INTERFACE, "CallTerminated",
-			DBUS_TYPE_INVALID);
-
-	if (hs->ph_number) {
-		g_free(hs->ph_number);
-		hs->ph_number = NULL;
-	}
-
-	if (hs->ring_timer) {
-		g_source_remove(hs->ring_timer);
-		hs->ring_timer = 0;
-	}
-
-	return headset_send(hs, "\r\nOK\n\r");
+	return 0;
 }
 
 static int cli_notification(struct audio_device *device, const char *buf)
@@ -742,6 +790,11 @@ static int cli_notification(struct audio_device *device, const char *buf)
 	return headset_send(hs, "\r\nOK\r\n");
 }
 
+int telephony_response_and_hold_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
 static int response_and_hold(struct audio_device *device, const char *buf)
 {
 	struct headset *hs = device->headset;
@@ -750,54 +803,57 @@ static int response_and_hold(struct audio_device *device, const char *buf)
 		return -EINVAL;
 
 	if (buf[7] == '=') {
-		if (telephony_response_and_hold_req(atoi(&buf[8]) < 0)) {
-			headset_send(hs, "\r\nERROR\r\n");
-			return 0;
-		}
-	} else if (ag.rh >= 0)
+		telephony_response_and_hold_req(device, atoi(&buf[8]) < 0);
+		return 0;
+	}
+
+	if (ag.rh >= 0)
 		headset_send(hs, "\r\n+BTRH:%d\r\n", ag.rh);
 
-	return headset_send(hs, "\r\nOK\n\r", ag.rh);
+	return headset_send(hs, "\r\nOK\r\n", ag.rh);
+}
+
+int telephony_last_dialed_number_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
 }
 
 static int last_dialed_number(struct audio_device *device, const char *buf)
 {
-	struct headset *hs = device->headset;
+	telephony_last_dialed_number_req(device);
 
-	ag.ev_buf_active = TRUE;
+	return 0;
+}
 
-	if (telephony_last_dialed_number_req() < 0) {
-		headset_send(hs, "\r\nERROR\r\n");
-		return 0;
-	}
-
-	flush_events();
-	ag.ev_buf_active = FALSE;
-
-	return headset_send(hs, "\r\nOK\n\r");
+int telephony_dial_number_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
 }
 
 static int dial_number(struct audio_device *device, const char *buf)
 {
-	struct headset *hs = device->headset;
+	char number[BUF_SIZE];
+	size_t buf_len;
 
-	ag.ev_buf_active = TRUE;
+	buf_len = strlen(buf);
 
-	if (telephony_dial_number_req(&buf[3]) < 0) {
-		headset_send(hs, "\r\nERROR\r\n");
-		return 0;
+	if (buf[buf_len - 1] != ';') {
+		debug("Rejecting non-voice call dial request");
+		return -EINVAL;
 	}
 
-	flush_events();
-	ag.ev_buf_active = FALSE;
+	memset(number, 0, sizeof(number));
+	strncpy(number, &buf[3], buf_len - 4);
 
-	return headset_send(hs, "\r\nOK\n\r");
+	telephony_dial_number_req(device, number);
+
+	return 0;
 }
 
-
 static int signal_gain_setting(struct audio_device *device, const char *buf)
 {
 	struct headset *hs = device->headset;
+	const char *property;
 	const char *name;
 	dbus_uint16_t gain;
 
@@ -818,12 +874,14 @@ static int signal_gain_setting(struct audio_device *device, const char *buf)
 		if (hs->sp_gain == gain)
 			goto ok;
 		name = "SpeakerGainChanged";
+		property = "SpeakerGain";
 		hs->sp_gain = gain;
 		break;
 	case HEADSET_GAIN_MICROPHONE:
 		if (hs->mic_gain == gain)
 			goto ok;
 		name = "MicrophoneGainChanged";
+		property = "MicrophoneGain";
 		hs->mic_gain = gain;
 		break;
 	default:
@@ -832,41 +890,133 @@ static int signal_gain_setting(struct audio_device *device, const char *buf)
 	}
 
 	g_dbus_emit_signal(device->conn, device->path,
-				    AUDIO_HEADSET_INTERFACE, name,
-				    DBUS_TYPE_UINT16, &gain,
-				    DBUS_TYPE_INVALID);
+				AUDIO_HEADSET_INTERFACE, name,
+				DBUS_TYPE_UINT16, &gain,
+				DBUS_TYPE_INVALID);
+
+	emit_property_changed(device->conn, device->path,
+				AUDIO_HEADSET_INTERFACE, property,
+				DBUS_TYPE_UINT16, &gain);
 
 ok:
 	return headset_send(hs, "\r\nOK\r\n");
 }
 
+int telephony_transmit_dtmf_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
 static int dtmf_tone(struct audio_device *device, const char *buf)
 {
-	struct headset *hs = device->headset;
-
 	if (strlen(buf) < 8) {
 		error("Too short string for DTMF tone");
 		return -EINVAL;
 	}
 
-	if (telephony_transmit_dtmf_req(buf[7]) < 0) {
-		headset_send(hs, "\r\nERROR\r\n");
-		return 0;
-	}
+	telephony_transmit_dtmf_req(device, buf[7]);
 
-	return headset_send(hs, "\r\nOK\n\r");
+	return 0;
+}
+
+int telephony_subscriber_number_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
 }
 
 static int subscriber_number(struct audio_device *device, const char *buf)
+{
+	telephony_subscriber_number_req(device);
+
+	return 0;
+}
+
+int telephony_list_current_calls_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int list_current_calls(struct audio_device *device, const char *buf)
+{
+	telephony_list_current_calls_req(device);
+
+	return 0;
+}
+
+static int extended_errors(struct audio_device *device, const char *buf)
 {
 	struct headset *hs = device->headset;
 
-	if (telephony_subscriber_number_req() < 0) {
-		headset_send(hs, "\r\nERROR\r\n");
-		return 0;
+	if (strlen(buf) < 9)
+		return -EINVAL;
+
+	if (buf[8] == '1') {
+		hs->cme_enabled = TRUE;
+		debug("CME errors enabled for headset %p", hs);
+	} else {
+		hs->cme_enabled = FALSE;
+		debug("CME errors disabled for headset %p", hs);
 	}
 
-	return headset_send(hs, "\r\nOK\n\r");
+	return headset_send(hs, "\r\nOK\r\n");
+}
+
+static int call_waiting_notify(struct audio_device *device, const char *buf)
+{
+	struct headset *hs = device->headset;
+
+	if (strlen(buf) < 9)
+		return -EINVAL;
+
+	if (buf[8] == '1') {
+		hs->cwa_enabled = TRUE;
+		debug("Call waiting notifiaction enabled for headset %p", hs);
+	} else {
+		hs->cwa_enabled = FALSE;
+		debug("Call waiting notification disabled for headset %p", hs);
+	}
+
+	return headset_send(hs, "\r\nOK\r\n");
+}
+
+int telephony_operator_selection_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+int telephony_call_hold_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+int telephony_operator_selection_ind(int mode, const char *oper)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	send_foreach_headset(active_devices, hfp_cmp, "\r\n+COPS:%d,0,%s\r\n",
+				mode, oper);
+	return 0;
+}
+
+static int operator_selection(struct audio_device *device, const char *buf)
+{
+	struct headset *hs = device->headset;
+
+	if (strlen(buf) < 8)
+		return -EINVAL;
+
+	switch (buf[7]) {
+	case '?':
+		telephony_operator_selection_req(device);
+		break;
+	case '=':
+		return headset_send(hs, "\r\nOK\r\n");
+	default:
+		return -EINVAL;
+	}
+
+	return 0;
 }
 
 static struct event event_callbacks[] = {
@@ -884,6 +1034,10 @@ static struct event event_callbacks[] = {
 	{ "AT+BLDN", last_dialed_number },
 	{ "AT+VTS", dtmf_tone },
 	{ "AT+CNUM", subscriber_number },
+	{ "AT+CLCC", list_current_calls },
+	{ "AT+CMEE", extended_errors },
+	{ "AT+CCWA", call_waiting_notify },
+	{ "AT+COPS", operator_selection },
 	{ 0 }
 };
 
@@ -1138,11 +1292,8 @@ static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 	return;
 
 failed_not_supported:
-	if (p->msg) {
+	if (p->msg)
 		error_not_supported(dev->conn, p->msg);
-		dbus_message_unref(p->msg);
-		p->msg = NULL;
-	}
 failed:
 	if (classes)
 		sdp_list_free(classes, free);
@@ -1328,26 +1479,25 @@ static DBusMessage *hs_connect(DBusConnection *conn, DBusMessage *msg,
 	return NULL;
 }
 
+static int cli_cmp(struct headset *hs)
+{
+	if (!hs->hfp_active)
+		return -1;
+
+	if (hs->cli_active)
+		return 0;
+	else
+		return -1;
+}
+
 static gboolean ring_timer_cb(gpointer data)
 {
-	struct audio_device *device = data;
-	struct headset *hs = device->headset;
-	int err;
+	send_foreach_headset(active_devices, NULL, "\r\nRING\r\n");
 
-	err = headset_send(hs, "\r\nRING\r\n");
-
-	if (err < 0)
-		error("Error while sending RING: %s (%d)",
-				strerror(-err), -err);
-
-	if (hs->cli_active && hs->ph_number) {
-		err = headset_send(hs, "\r\n+CLIP:\"%s\",%d\r\n",
-					hs->ph_number, hs->type);
-
-		if (err < 0)
-			error("Error while sending CLIP: %s (%d)",
-					strerror(-err), -err);
-	}
+	if (ag.number)
+		send_foreach_headset(active_devices, cli_cmp,
+					"\r\n+CLIP:\"%s\",%d\r\n",
+					ag.number, ag.number_type);
 
 	return TRUE;
 }
@@ -1369,7 +1519,7 @@ static DBusMessage *hs_ring(DBusConnection *conn, DBusMessage *msg,
 	if (!reply)
 		return NULL;
 
-	if (hs->ring_timer) {
+	if (ag.ring_timer) {
 		debug("IndicateCall received when already indicating");
 		goto done;
 	}
@@ -1381,18 +1531,8 @@ static DBusMessage *hs_ring(DBusConnection *conn, DBusMessage *msg,
 						"%s", strerror(-err));
 	}
 
-	if (hs->cli_active && hs->ph_number) {
-		err = headset_send(hs, "\r\n+CLIP:\"%s\",%d\r\n",
-					hs->ph_number, hs->type);
-		if (err < 0) {
-			dbus_message_unref(reply);
-			return g_dbus_create_error(msg, ERROR_INTERFACE
-						".Failed", "%s",
-						strerror(-err));
-		}
-	}
-
-	hs->ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb, device);
+	ring_timer_cb(NULL);
+	ag.ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb, NULL);
 
 done:
 	return reply;
@@ -1415,9 +1555,9 @@ static DBusMessage *hs_cancel_call(DBusConnection *conn,
 	if (!reply)
 		return NULL;
 
-	if (hs->ring_timer) {
-		g_source_remove(hs->ring_timer);
-		hs->ring_timer = 0;
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
 	} else
 		debug("Got CancelCall method call but no call is active");
 
@@ -1519,12 +1659,12 @@ static DBusMessage *hs_get_mic_gain(DBusConnection *conn,
 
 static DBusMessage *hs_set_gain(DBusConnection *conn,
 				DBusMessage *msg,
-				void *data, char type)
+				void *data, uint16_t gain,
+				char type)
 {
 	struct audio_device *device = data;
 	struct headset *hs = device->headset;
 	DBusMessage *reply;
-	dbus_uint16_t gain;
 	int err;
 
 	if (hs->state < HEADSET_STATE_CONNECTED)
@@ -1532,10 +1672,6 @@ static DBusMessage *hs_set_gain(DBusConnection *conn,
 						".NotConnected",
 						"Device not Connected");
 
-	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT16, &gain,
-				DBUS_TYPE_INVALID))
-		return NULL;
-
 	if (gain > 15)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
 						".InvalidArgument",
@@ -1559,17 +1695,17 @@ done:
 	if (type == HEADSET_GAIN_SPEAKER) {
 		hs->sp_gain = gain;
 		g_dbus_emit_signal(conn, device->path,
-						AUDIO_HEADSET_INTERFACE,
-						"SpeakerGainChanged",
-						DBUS_TYPE_UINT16, &gain,
-						DBUS_TYPE_INVALID);
+					AUDIO_HEADSET_INTERFACE,
+					"SpeakerGainChanged",
+					DBUS_TYPE_UINT16, &gain,
+					DBUS_TYPE_INVALID);
 	} else {
 		hs->mic_gain = gain;
 		g_dbus_emit_signal(conn, device->path,
-						AUDIO_HEADSET_INTERFACE,
-						"MicrophoneGainChanged",
-						DBUS_TYPE_UINT16, &gain,
-						DBUS_TYPE_INVALID);
+					AUDIO_HEADSET_INTERFACE,
+					"MicrophoneGainChanged",
+					DBUS_TYPE_UINT16, &gain,
+					DBUS_TYPE_INVALID);
 	}
 
 	return reply;
@@ -1579,16 +1715,113 @@ static DBusMessage *hs_set_speaker_gain(DBusConnection *conn,
 					DBusMessage *msg,
 					void *data)
 {
-	return hs_set_gain(conn, msg, data, HEADSET_GAIN_SPEAKER);
+	uint16_t gain;
+
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT16, &gain,
+				DBUS_TYPE_INVALID))
+		return NULL;
+
+	return hs_set_gain(conn, msg, data, gain, HEADSET_GAIN_SPEAKER);
 }
 
 static DBusMessage *hs_set_mic_gain(DBusConnection *conn,
 					DBusMessage *msg,
 					void *data)
 {
-	return hs_set_gain(conn, msg, data, HEADSET_GAIN_MICROPHONE);
+	uint16_t gain;
+
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT16, &gain,
+				DBUS_TYPE_INVALID))
+		return NULL;
+
+	return hs_set_gain(conn, msg, data, gain, HEADSET_GAIN_MICROPHONE);
 }
 
+static DBusMessage *hs_get_properties(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	struct audio_device *device = data;
+	DBusMessage *reply;
+	DBusMessageIter iter;
+	DBusMessageIter dict;
+	gboolean value;
+
+	reply = dbus_message_new_method_return(msg);
+	if (!reply)
+		return NULL;
+
+	dbus_message_iter_init_append(reply, &iter);
+
+	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
+			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
+			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+
+
+	/* Playing */
+	value = (device->headset->state == HEADSET_STATE_PLAYING);
+	dict_append_entry(&dict, "Playing", DBUS_TYPE_BOOLEAN, &value);
+
+	/* Connected */
+	value = (device->headset->state >= HEADSET_STATE_CONNECTED);
+	dict_append_entry(&dict, "Connected", DBUS_TYPE_BOOLEAN, &value);
+
+	if (!value)
+		goto done;
+
+	/* SpeakerGain */
+	dict_append_entry(&dict, "SpeakerGain",
+				DBUS_TYPE_UINT16, &device->headset->sp_gain);
+
+	/* MicrophoneGain */
+	dict_append_entry(&dict, "MicrophoneGain",
+				DBUS_TYPE_UINT16, &device->headset->mic_gain);
+
+done:
+	dbus_message_iter_close_container(&iter, &dict);
+
+	return reply;
+}
+
+static DBusMessage *hs_set_property(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *property;
+	DBusMessageIter iter;
+	DBusMessageIter sub;
+	uint16_t gain;
+
+	if (!dbus_message_iter_init(msg, &iter))
+		return invalid_args(msg);
+
+	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
+		return invalid_args(msg);
+
+	dbus_message_iter_get_basic(&iter, &property);
+	dbus_message_iter_next(&iter);
+
+	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
+		return invalid_args(msg);
+	dbus_message_iter_recurse(&iter, &sub);
+
+	if (g_str_equal("SpeakerGain", property)) {
+		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_UINT16)
+			return invalid_args(msg);
+
+		dbus_message_iter_get_basic(&sub, &gain);
+		return hs_set_gain(conn, msg, data, gain,
+					HEADSET_GAIN_SPEAKER);
+	} else if (g_str_equal("MicrophoneGain", property)) {
+		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_UINT16)
+			return invalid_args(msg);
+
+		dbus_message_iter_get_basic(&sub, &gain);
+		return hs_set_gain(conn, msg, data, gain,
+					HEADSET_GAIN_MICROPHONE);
+	}
+
+	return invalid_args(msg);
+}
 static GDBusMethodTable headset_methods[] = {
 	{ "Connect",		"",	"",	hs_connect,
 						G_DBUS_METHOD_FLAG_ASYNC },
@@ -1599,28 +1832,36 @@ static GDBusMethodTable headset_methods[] = {
 	{ "Play",		"",	"",	hs_play,
 						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Stop",		"",	"",	hs_stop },
-	{ "IsPlaying",		"",	"b",	hs_is_playing },
-	{ "GetSpeakerGain",	"",	"q",	hs_get_speaker_gain },
-	{ "GetMicrophoneGain",	"",	"q",	hs_get_mic_gain },
-	{ "SetSpeakerGain",	"q",	"",	hs_set_speaker_gain },
-	{ "SetMicrophoneGain",	"q",	"",	hs_set_mic_gain },
+	{ "IsPlaying",		"",	"b",	hs_is_playing,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "GetSpeakerGain",	"",	"q",	hs_get_speaker_gain,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "GetMicrophoneGain",	"",	"q",	hs_get_mic_gain,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "SetSpeakerGain",	"q",	"",	hs_set_speaker_gain,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "SetMicrophoneGain",	"q",	"",	hs_set_mic_gain,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "GetProperties",	"",	"a{sv}",hs_get_properties },
+	{ "SetProperty",	"sv",	"",	hs_set_property },
 	{ NULL, NULL, NULL, NULL }
 };
 
 static GDBusSignalTable headset_signals[] = {
-	{ "Connected",			""	},
-	{ "Disconnected",		""	},
+	{ "Connected",			"",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
+	{ "Disconnected",		"",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
 	{ "AnswerRequested",		""	},
-	{ "Stopped",			""	},
-	{ "Playing",			""	},
-	{ "SpeakerGainChanged",		"q"	},
-	{ "MicrophoneGainChanged",	"q"	},
+	{ "Stopped",			"",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
+	{ "Playing",			"",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
+	{ "SpeakerGainChanged",		"q",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
+	{ "MicrophoneGainChanged",	"q",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
 	{ "CallTerminated",		""	},
+	{ "PropertyChanged",		"sv"	},
 	{ NULL, NULL }
 };
 
-static void headset_set_channel(struct headset *headset, sdp_record_t *record,
-				uint16_t svc)
+static void headset_set_channel(struct headset *headset,
+				const sdp_record_t *record, uint16_t svc)
 {
 	int ch;
 	sdp_list_t *protos;
@@ -1644,9 +1885,15 @@ static void headset_set_channel(struct headset *headset, sdp_record_t *record,
 		error("Unable to get RFCOMM channel from Headset record");
 }
 
-void headset_update(struct audio_device *dev, sdp_record_t *record, uint16_t svc)
+void headset_update(struct audio_device *dev, uint16_t svc,
+			const char *uuidstr)
 {
 	struct headset *headset = dev->headset;
+	const sdp_record_t *record;
+
+	record = btd_device_get_record(dev->btd_dev, uuidstr);
+	if (!record)
+		return;
 
 	switch (svc) {
 	case HANDSFREE_SVCLASS_ID:
@@ -1708,6 +1955,12 @@ static void headset_free(struct audio_device *dev)
 static void path_unregister(void *data)
 {
 	struct audio_device *dev = data;
+	struct headset *hs = dev->headset;
+
+	if (hs->state > HEADSET_STATE_DISCONNECTED) {
+		debug("Headset unregistered while device was connected!");
+		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
+	}
 
 	info("Unregistered interface %s on path %s",
 		AUDIO_HEADSET_INTERFACE, dev->path);
@@ -1721,10 +1974,11 @@ void headset_unregister(struct audio_device *dev)
 		AUDIO_HEADSET_INTERFACE);
 }
 
-struct headset *headset_init(struct audio_device *dev, sdp_record_t *record,
-				uint16_t svc)
+struct headset *headset_init(struct audio_device *dev, uint16_t svc,
+				const char *uuidstr)
 {
 	struct headset *hs;
+	const sdp_record_t *record;
 
 	hs = g_new0(struct headset, 1);
 	hs->rfcomm_ch = -1;
@@ -1733,8 +1987,8 @@ struct headset *headset_init(struct audio_device *dev, sdp_record_t *record,
 	hs->search_hfp = server_is_enabled(&dev->src, HANDSFREE_SVCLASS_ID);
 	hs->hfp_active = FALSE;
 	hs->cli_active = FALSE;
-	hs->ph_number = NULL;
 
+	record = btd_device_get_record(dev->btd_dev, uuidstr);
 	if (!record)
 		goto register_iface;
 
@@ -1831,8 +2085,6 @@ gboolean headset_cancel_stream(struct audio_device *dev, unsigned int id)
 	if (p->callbacks || p->msg)
 		return TRUE;
 
-	pending_connect_finalize(dev);
-
 	if (hs->auto_dc) {
 		if (hs->rfcomm)
 			hs->dc_timer = g_timeout_add(DC_TIMEOUT,
@@ -1920,16 +2172,11 @@ int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
 	return 0;
 }
 
-int headset_close_rfcomm(struct audio_device *dev)
+static int headset_close_rfcomm(struct audio_device *dev)
 {
 	struct headset *hs = dev->headset;
 	GIOChannel *rfcomm = hs->tmp_rfcomm ? hs->tmp_rfcomm : hs->rfcomm;
 
-	if (hs->ring_timer) {
-		g_source_remove(hs->ring_timer);
-		hs->ring_timer = 0;
-	}
-
 	if (rfcomm) {
 		g_io_channel_close(rfcomm);
 		g_io_channel_unref(rfcomm);
@@ -1972,50 +2219,68 @@ void headset_set_authorized(struct audio_device *dev)
 void headset_set_state(struct audio_device *dev, headset_state_t state)
 {
 	struct headset *hs = dev->headset;
+	gboolean value;
 
 	if (hs->state == state)
 		return;
 
 	switch (state) {
 	case HEADSET_STATE_DISCONNECTED:
+		value = FALSE;
 		close_sco(dev);
 		headset_close_rfcomm(dev);
 		g_dbus_emit_signal(dev->conn, dev->path,
-						AUDIO_HEADSET_INTERFACE,
-						"Disconnected",
-						DBUS_TYPE_INVALID);
-		telephony_event_reporting_req(0);
-		if (dev == active_telephony_device)
-			active_telephony_device = NULL;
+					AUDIO_HEADSET_INTERFACE,
+					"Disconnected",
+					DBUS_TYPE_INVALID);
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "Connected",
+					DBUS_TYPE_BOOLEAN, &value);
+		telephony_device_disconnected(dev);
+		active_devices = g_slist_remove(active_devices, dev);
 		break;
 	case HEADSET_STATE_CONNECT_IN_PROGRESS:
 		break;
 	case HEADSET_STATE_CONNECTED:
 		close_sco(dev);
 		if (hs->state < state) {
+			value = TRUE;
 			g_dbus_emit_signal(dev->conn, dev->path,
 						AUDIO_HEADSET_INTERFACE,
 						"Connected",
 						DBUS_TYPE_INVALID);
-			if (!active_telephony_device)
-				active_telephony_device = dev;
+			emit_property_changed(dev->conn, dev->path,
+						AUDIO_HEADSET_INTERFACE,
+						"Connected",
+						DBUS_TYPE_BOOLEAN, &value);
+			active_devices = g_slist_append(active_devices, dev);
+			telephony_device_connected(dev);
 		} else if (hs->state == HEADSET_STATE_PLAYING) {
+			value = FALSE;
 			g_dbus_emit_signal(dev->conn, dev->path,
 						AUDIO_HEADSET_INTERFACE,
 						"Stopped",
 						DBUS_TYPE_INVALID);
+			emit_property_changed(dev->conn, dev->path,
+						AUDIO_HEADSET_INTERFACE,
+						"Playing",
+						DBUS_TYPE_BOOLEAN, &value);
 		}
 		break;
 	case HEADSET_STATE_PLAY_IN_PROGRESS:
 		break;
 	case HEADSET_STATE_PLAYING:
+		value = TRUE;
 		hs->sco_id = g_io_add_watch(hs->sco,
 					G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 					(GIOFunc) sco_cb, dev);
 
 		g_dbus_emit_signal(dev->conn, dev->path,
-						AUDIO_HEADSET_INTERFACE,
-						"Playing", DBUS_TYPE_INVALID);
+					AUDIO_HEADSET_INTERFACE, "Playing",
+					DBUS_TYPE_INVALID);
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "Playing",
+					DBUS_TYPE_BOOLEAN, &value);
 
 		if (hs->sp_gain >= 0)
 			headset_send(hs, "\r\n+VGS=%u\r\n", hs->sp_gain);
@@ -2114,108 +2379,81 @@ int headset_get_sco_fd(struct audio_device *dev)
 
 int telephony_event_ind(int index)
 {
-	struct headset *hs;
-
-	if (!active_telephony_device)
+	if (!active_devices)
 		return -ENODEV;
 
-	hs = active_telephony_device->headset;
-
-	if (!hs->hfp_active)
-		return -EINVAL;
-
 	if (!ag.er_ind) {
 		debug("telephony_report_event called but events are disabled");
 		return -EINVAL;
 	}
 
-	if (ag.ev_buf_active)
-		return buffer_event(index);
-
-	return headset_send(hs, "\r\n+CIEV:%d,%d\r\n", index + 1,
+	send_foreach_headset(active_devices, hfp_cmp,
+				"\r\n+CIEV:%d,%d\r\n", index + 1,
 				ag.indicators[index].val);
+
+	return 0;
 }
 
 int telephony_response_and_hold_ind(int rh)
 {
-	struct headset *hs;
-
-	if (!active_telephony_device)
+	if (!active_devices)
 		return -ENODEV;
 
-	hs = active_telephony_device->headset;
-
-	if (!hs->hfp_active)
-		return -EINVAL;
-
 	ag.rh = rh;
 
 	/* If we aren't in any response and hold state don't send anything */
 	if (ag.rh < 0)
 		return 0;
 
-	return headset_send(hs, "\r\n+BTRH:%d\r\n", ag.rh);
+	send_foreach_headset(active_devices, hfp_cmp, "\r\n+BTRH:%d\r\n",
+				ag.rh);
+
+	return 0;
 }
 
-int telephony_calling_started_ind(const char *number)
+int telephony_incoming_call_ind(const char *number, int type)
 {
-	struct headset *hs;
-
-	if (!active_telephony_device)
+	if (!active_devices)
 		return -ENODEV;
 
-	hs = active_telephony_device->headset;
-
-	if (hs->ring_timer) {
-		debug("telephony_notify_call: already calling");
+	if (ag.ring_timer) {
+		debug("telephony_incoming_call_ind: already calling");
 		return -EBUSY;
 	}
 
-	if (hs->ph_number) {
-		g_free(hs->ph_number);
-		hs->ph_number = NULL;
-	}
+	g_free(ag.number);
+	ag.number = g_strdup(number);
+	ag.number_type = type;
 
-	if (number)
-		hs->ph_number = g_strdup(number);
-
-	headset_send(hs, "\r\nRING\r\n");
-
-	if (hs->cli_active && hs->ph_number)
-		headset_send(hs, "\r\n+CLIP:\"%s\",%d\r\n",
-				hs->ph_number, hs->type);
-
-	hs->ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb,
-					active_telephony_device);
+	ring_timer_cb(NULL);
+	ag.ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb, NULL);
 
 	return 0;
 }
 
 int telephony_calling_stopped_ind(void)
 {
-	struct headset *hs;
-
-	if (!active_telephony_device)
+	if (!active_devices)
 		return -ENODEV;
 
-	hs = active_telephony_device->headset;
-
-	if (!hs->ring_timer)
+	if (!ag.ring_timer)
 		return -EINVAL;
 
-	g_source_remove(hs->ring_timer);
-	hs->ring_timer = 0;
+	g_source_remove(ag.ring_timer);
+	ag.ring_timer = 0;
 
 	return 0;
 }
 
 int telephony_ready_ind(uint32_t features,
-			const struct indicator *indicators, int rh)
+			const struct indicator *indicators, int rh,
+			const char *chld)
 {
 	ag.telephony_ready = TRUE;
 	ag.features = features;
 	ag.indicators = indicators;
 	ag.rh = rh;
+	ag.chld = g_strdup(chld);
 
 	debug("Telephony plugin initialized");
 
@@ -2223,3 +2461,57 @@ int telephony_ready_ind(uint32_t features,
 
 	return 0;
 }
+
+int telephony_list_current_call_ind(int idx, int dir, int status, int mode,
+					int mprty, const char *number,
+					int type)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	if (number)
+		send_foreach_headset(active_devices, hfp_cmp,
+					"\r\n+CLCC:%d,%d,%d,%d,%d,%s,%d\r\n",
+					idx, dir, status, mode, mprty,
+					number, type);
+	else
+		send_foreach_headset(active_devices, hfp_cmp,
+					"\r\n+CLCC:%d,%d,%d,%d,%d\r\n",
+					idx, dir, status, mode, mprty);
+
+	return 0;
+}
+
+int telephony_subscriber_number_ind(const char *number, int type, int service)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	send_foreach_headset(active_devices, hfp_cmp,
+				"\r\n+CNUM:,%s,%d,,%d\r\n",
+				number, type, service);
+
+	return 0;
+}
+
+static int cwa_cmp(struct headset *hs)
+{
+	if (!hs->hfp_active)
+		return -1;
+
+	if (hs->cwa_enabled)
+		return 0;
+	else
+		return -1;
+}
+
+int telephony_call_waiting_ind(const char *number, int type)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	send_foreach_headset(active_devices, cwa_cmp, "\r\n+CCWA:%s,%d\r\n",
+				number, type);
+
+	return 0;
+}
